from urllib.request import Request, urlopen

import requests

from ..tools.logger import Logger


def download_web(url, filename):
    """ 下载网页 """
    try:
        headers = {'User-Agent':'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:23.0) Gecko/20100101 Firefox/23.0'}
        req = Request(url=url, headers=headers)
        content = urlopen(req).read().decode("utf-8")
        if len(content) > 0:
            with open(filename, 'w') as fw:
                fw.write(content)
        return True
    except Exception as e:
        print(e)
        return False

def get_web_text(url):
    """获取网页文本"""
    try:
        headers = {'User-Agent':'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:23.0) Gecko/20100101 Firefox/23.0'}
        req = Request(url=url, headers=headers)
        content = urlopen(req).read().decode("utf-8")
        return content
    except Exception as e:
        print(e)
        return ""

def get_github_star(url):
    """获取github标星数量 以及 上次提交时间 很容易被封"""
    req = Request(url=url)
    content = urlopen(req).read().decode("utf-8")
    star_num = content.split("users starred this repository")[0].split("\"")[-1].strip()
    star_num = int(star_num)

    last_time = content.split('<relative-time datetime="')[1].split("T")[0]
    return star_num, last_time

def youdao_translate(word):
    """ 使用有道词典翻译 """

    try:
        url = f'http://dict.youdao.com/w/{word}/#keyfrom=dict2.top'
        text_list = requests.get(url=url).text.split('<ul>')[1].split('</ul>')[0].split('<li>')[1:]
        item_list = []
        for text in text_list:
            item = text.split('</li>')[0].strip()
            if '人名' in item:
                continue
            item_list.append(item)
        return item_list
    
    except Exception:
        Logger.warn('没有翻译结果 或者 出现网络故障!')
        return []


def parse_google_text(web_text):
    """ 解析谷歌学术页面源码 """

    paper_list = []
    title_set = set()
    for text in web_text.split('data-cid'):
        segs = text.split('data-clk-atid')
        if len(segs) != 3:
            continue
        title = segs[-1].split('</a>')[0].split('">')[1].replace('<b>', ' ').replace('</b>', ' ').strip()

        if title in title_set:
            continue
        title_set.add(title)

        pdf_url = text.split('tabindex')[-1].split('<a href="')[1].split('"')[0]

        cite_num = 0
        if '被引用次数：' in text:
            cite_num = text.split('被引用次数：')[1].split('</a>')[0]
            cite_num = int(cite_num)

        paper_list.append(
            {
                "title": title,
                "cite_num": cite_num,
                "pdf_url": pdf_url,
            }
        )

    return paper_list



import time


class Timer(object):
    def __init__(self) -> None:

        self._total_time = 0.0
        self._count = 0
        self._start = None

    def tic(self) -> None:
        self._start = time.time()

    def toc(self) -> float:
        t = 0.0
        if self._start is not None:
            t = time.time() - self._start
            self._start = None

        self._total_time += t
        self._count += 1
        return t

    def total(self) -> float:
        """total time"""
        return self._total_time

    def count(self) -> int:
        return self._count

    def fps(self) -> float:
        if self._count == 0:
            return 0.0
        return self._count / self._total_time



from .track_base import BaseTrackObj
from ..vis.plot_kps import imshow


class ImageTrackObj(BaseTrackObj):
    def __init__(self, img, name=None) -> None:
        super().__init__()

        self.img = img
        self.set_name(name)

    def data(self):
        return self.img

    def show(self):
        imshow("img", self.img, wait_ms=0)



"""
Copyright (c) Microsoft Corporation.
Licensed under the MIT license.

Definition of TSV class
"""
# copy from https://github.com/microsoft/MeshTransformer/tree/main/metro/utils/tsv_file.py

import logging
import os
import os.path as op


def generate_lineidx(filein, idxout):
    idxout_tmp = idxout + ".tmp"
    with open(filein, "r") as tsvin, open(idxout_tmp, "w") as tsvout:
        fsize = os.fstat(tsvin.fileno()).st_size
        fpos = 0
        while fpos != fsize:
            tsvout.write(str(fpos) + "\n")
            tsvin.readline()
            fpos = tsvin.tell()
    os.rename(idxout_tmp, idxout)


class TSVFileReader(object):
    def __init__(self, tsv_file):
        self.tsv_file = tsv_file
        self.lineidx = op.splitext(tsv_file)[0] + ".lineidx"
        self._fp = None
        self._lineidx = None
        # the process always keeps the process which opens the file.
        # If the pid is not equal to the currrent pid, we will re-open the file.
        self.pid = None
        # generate lineidx if not exist
        if not op.isfile(self.lineidx):
            generate_lineidx(self.tsv_file, self.lineidx)

    def __del__(self):
        if self._fp:
            self._fp.close()

    def __str__(self):
        return "TSVFile(tsv_file='{}')".format(self.tsv_file)

    def __repr__(self):
        return str(self)

    def num_rows(self):
        self._ensure_lineidx_loaded()
        return len(self._lineidx)

    def seek(self, idx):
        self._ensure_tsv_opened()
        self._ensure_lineidx_loaded()
        try:
            pos = self._lineidx[idx]
        except:
            logging.info("{}-{}".format(self.tsv_file, idx))
            raise
        self._fp.seek(pos)
        return [s.strip() for s in self._fp.readline().split("\t")]

    def __getitem__(self, index):
        return self.seek(index)

    def __len__(self):
        return self.num_rows()

    def _ensure_lineidx_loaded(self):
        if self._lineidx is None:
            logging.info("loading lineidx: {}".format(self.lineidx))
            with open(self.lineidx, "r") as fp:
                self._lineidx = [int(i.strip()) for i in fp.readlines()]

    def _ensure_tsv_opened(self):
        if self._fp is None:
            self._fp = open(self.tsv_file, "r")
            self.pid = os.getpid()

        if self.pid != os.getpid():
            logging.info(
                "re-open {} because the process id changed".format(self.tsv_file)
            )
            self._fp = open(self.tsv_file, "r")
            self.pid = os.getpid()


class TSVFileWriter(object):
    def __init__(self, tsv_file):
        self.tsv_file = tsv_file
        self._fp = None

    def writeline(self, text):
        self._ensure_tsv_opened()
        if isinstance(text, str):
            self._fp.write(text + "\n")
        elif isinstance(text, list):
            text = "\t".join(text)
            self._fp.write(text + "\n")

    def __del__(self):
        self.close()
    
    def _ensure_tsv_opened(self):
        if self._fp is None:
            self._fp = open(self.tsv_file, "w")
    
    def close(self):
        if self._fp:
            self._fp.close()



import importlib
import os.path as osp
import shutil

from omegaconf import OmegaConf

from ..tools.logger import Logger


def get_obj_from_str(string, reload=False):
    module, cls = string.rsplit(".", 1)
    if reload:
        module_imp = importlib.import_module(module)
        importlib.reload(module_imp)
    return getattr(importlib.import_module(module, package=None), cls)


def instantiate_from_config(config):
    if not "target" in config:
        raise KeyError("Expected key `target` to instantiate.")
    
    if "params" in config:
        return get_obj_from_str(config["target"])(**config.get("params", dict()))
    else:
        return get_obj_from_str(config["target"])()


def load_config(filename):
    """加载参数文件"""

    filename = osp.abspath(filename)
    config = OmegaConf.load(filename)

    Logger.info(f"load_config: {filename}")
    # merge config
    if hasattr(config, "base"):
        base_filename = osp.join(osp.dirname(filename), config.base)
        delattr(config, "base")

        base_config = load_config(base_filename)
        config = OmegaConf.merge(base_config, config)
    return config


def save_config(filename, config):
    OmegaConf.save(config, filename)
    Logger.info(f"save_config: {filename}")


def make_config():
    src = osp.join(osp.dirname(__file__), "template.yaml")
    dst = "template.yaml"
    shutil.copy(src, dst)
    Logger.info(f"create: {dst}")


def get_log_dir(config_filename):
    return config_filename.replace("configs", "logs").replace(".yaml", "")

def get_checkpoint_dir(config_filename):
    return osp.join(get_log_dir(config_filename), "checkpoint")

def omegaconf_to_dict(cfg):
    d = dict()
    for key, value in cfg.items():
        d[key] = value
    return d



import pytorch_lightning as pl
import torch  # do not remove
import torch.nn as nn
from torchvision.utils import make_grid

from ..tools.logger import Logger


class BaseModel(nn.Module):
    def __init__(self):
        super().__init__()

    def run_train(self, batch):
        raise NotImplementedError("not implemented!")

    def run_validation(self, batch):
        raise NotImplementedError("not implemented!")

    def training_epoch_end(self):
        pass

    def validation_epoch_end(self):
        pass


class LightningModel(pl.LightningModule):
    def __init__(self, model, optimizer_cfg, scheduler_cfg) -> None:
        super().__init__()

        self.model = model
        self.optimizer_cfg = optimizer_cfg
        self.scheduler_cfg = scheduler_cfg
        self.log_img_cnt = 0

    def configure_optimizers(self):

        # optimizer
        optimizer_cls = f"torch.optim.{self.optimizer_cfg.target}"
        Logger.info(f"optimizer:{optimizer_cls}")
        optimizer = eval(optimizer_cls)(
            filter(lambda p: p.requires_grad, self.model.parameters()),
            **self.optimizer_cfg.params,
        )

        # scheduler
        scheduler_cls = f"torch.optim.lr_scheduler.{self.scheduler_cfg.target}"
        Logger.info(f"scheduler:{scheduler_cls}")
        scheduler = eval(scheduler_cls)(optimizer, **self.scheduler_cfg.params)

        return ({"optimizer": optimizer, "lr_scheduler": scheduler},)

    def log_img(self, img_dict_list):
        self.log_img_cnt += 1

        if img_dict_list is not None:
            tensorboard = self.logger.experiment[0]
            for img_dict in img_dict_list:
                name = img_dict["name"]
                img = img_dict["img"]
                nrow = img_dict["nrow"] if "nrow" in img_dict else 8
                tensorboard.add_image(name, make_grid(img, nrow=nrow), self.log_img_cnt)

    def training_step(self, batch, batch_idx):
        # run_train
        if not hasattr(self.model, "current_epoch"):
            setattr(self.model, "current_epoch", self.current_epoch)
        else:
            self.model.current_epoch = self.current_epoch

        total_loss, loss_dict, img_dict_list = self.model.run_train(batch)
        self.log_img(img_dict_list)

        for name, value in loss_dict.items():
            self.log(f"train/{name}", value, on_step=True, on_epoch=False)

        for name, value in loss_dict.items():
            self.log(f"train_epoch/{name}", value, on_step=False, on_epoch=True)

        return total_loss

    def validation_step(self, batch, batch_idx):
        # run_eval
        batch_size = batch[list(batch.keys())[0]].shape[0]
        total_loss, loss_dict, img_dict_list = self.model.run_validation(batch)
        self.log_img(img_dict_list)

        for name, value in loss_dict.items():
            self.log(f"eval/{name}", value, on_epoch=True)

        self.log(
            "eval_loss",
            total_loss,
            prog_bar=True,
        )
        loss_dict["_batch_size"] = batch_size
        return loss_dict

    def validation_epoch_end(self, loss_list):
        if isinstance(self.model, BaseModel):
            self.model.validation_epoch_end()
        total_loss = dict()

        total_cnt = 0
        for loss_dict in loss_list:
            batch_size = loss_dict["_batch_size"]
            total_cnt += batch_size

            for name, loss in loss_dict.items():
                if name == "_batch_size":
                    continue

                if name not in total_loss:
                    total_loss[name] = 0
                total_loss[name] += loss * batch_size

        text = f"epoch={self.current_epoch:03d}, "
        for name in total_loss:
            loss = total_loss[name] / total_cnt
            text += f"{name}={loss:.5f}, "

        Logger.info(text[:-2])

    def training_epoch_end(self, loss_list) -> None:
        if isinstance(self.model, BaseModel):
            self.model.training_epoch_end()



import torch
import numpy as np

__all__ = ["mpjpe", "p_mpjpe", "n_mpjpe", "weighted_mpjpe", "mean_per_vertex_error"]


def mpjpe(predicted, target):
    """
    Mean per-joint position error (i.e. mean Euclidean distance),
    often referred to as "Protocol #1" in many papers.
    """
    assert predicted.shape == target.shape
    return torch.mean(torch.norm(predicted - target, dim=-1))


def weighted_mpjpe(predicted, target, w):
    """
    Weighted mean per-joint position error (i.e. mean Euclidean distance)
    """
    assert predicted.shape == target.shape
    assert w.shape[0] == predicted.shape[0]
    return torch.mean(w * torch.norm(predicted - target, dim=-1))

def mean_per_vertex_error(pred, gt, has_smpl):
    """
    Compute mPVE
    """
    pred = pred[has_smpl == 1]
    gt = gt[has_smpl == 1]
    with torch.no_grad():
        error = torch.sqrt( ((pred - gt) ** 2).sum(dim=-1)).mean(dim=-1).cpu().numpy()
        return error

def p_mpjpe(predicted, target):
    """
    Pose error: MPJPE after rigid alignment (scale, rotation, and translation),
    often referred to as "Protocol #2" in many papers.
    """
    assert predicted.shape == target.shape
    assert len(predicted.shape) == 3
    assert type(predicted) is np.ndarray
    assert type(target) is np.ndarray

    muX = np.mean(target, axis=1, keepdims=True)
    muY = np.mean(predicted, axis=1, keepdims=True)

    X0 = target - muX
    Y0 = predicted - muY

    normX = np.sqrt(np.sum(X0 ** 2, axis=(1, 2), keepdims=True))
    normY = np.sqrt(np.sum(Y0 ** 2, axis=(1, 2), keepdims=True))

    X0 /= normX
    Y0 /= normY

    H = np.matmul(X0.transpose(0, 2, 1), Y0)
    U, s, Vt = np.linalg.svd(H)
    V = Vt.transpose(0, 2, 1)
    R = np.matmul(V, U.transpose(0, 2, 1))

    # Avoid improper rotations (reflections), i.e. rotations with det(R) = -1
    sign_detR = np.sign(np.expand_dims(np.linalg.det(R), axis=1))
    V[:, :, -1] *= sign_detR
    s[:, -1] *= sign_detR.flatten()
    R = np.matmul(V, U.transpose(0, 2, 1))  # Rotation

    tr = np.expand_dims(np.sum(s, axis=1, keepdims=True), axis=2)

    a = tr * normX / normY  # Scale
    t = muX - a * np.matmul(muY, R)  # Translation

    # Perform rigid transformation on the input
    predicted_aligned = a * np.matmul(predicted, R) + t

    # Return MPJPE
    return np.mean(
        np.linalg.norm(predicted_aligned - target, axis=len(target.shape) - 1)
    )


def n_mpjpe(predicted, target):
    """
    Normalized MPJPE (scale only), adapted from:
    https://github.com/hrhodin/UnsupervisedGeometryAwareRepresentationLearning/blob/master/losses/poses.py

    https://github.com/facebookresearch/VideoPose3D/blob/master/common/loss.py
    """
    assert predicted.shape == target.shape

    # only support len(predicted.shape)=4
    # norm_predicted = torch.mean(
    # torch.sum(predicted**2, dim=3, keepdim=True), dim=2, keepdim=True)
    # norm_target = torch.mean(
    # torch.sum(target*predicted, dim=3, keepdim=True), dim=2, keepdim=True)

    # can support len(predicted.shape) >= 2
    norm_predicted = torch.mean(
        torch.sum(predicted ** 2, dim=-1, keepdim=True), dim=-2, keepdim=True
    )
    norm_target = torch.mean(
        torch.sum(target * predicted, dim=-1, keepdim=True), dim=-2, keepdim=True
    )

    scale = norm_target / norm_predicted
    return mpjpe(scale * predicted, target)


def mean_velocity_error(predicted, target):
    """
    Mean per-joint velocity error (i.e. mean Euclidean distance of the 1st derivative)
    """
    assert predicted.shape == target.shape

    velocity_predicted = np.diff(predicted, axis=0)
    velocity_target = np.diff(target, axis=0)

    return np.mean(
        np.linalg.norm(velocity_predicted - velocity_target, axis=len(target.shape) - 1)
    )


def estimate_pose(points_static, points_to_transform):
    # pdb.set_trace()
    p0 = np.copy(points_static).T
    p1 = np.copy(points_to_transform).T

    t0 = -np.mean(p0, axis=1).reshape(3, 1)
    t1 = -np.mean(p1, axis=1).reshape(3, 1)

    p0c = p0 + t0
    p1c = p1 + t1

    covariance_matrix = p0c.dot(p1c.T)
    U, S, V = np.linalg.svd(covariance_matrix)
    R = U.dot(V)
    if np.linalg.det(R) < 0:
        R[:, 2] *= -1

    rms_d0 = np.sqrt(np.mean(np.linalg.norm(p0c, axis=0) ** 2))
    rms_d1 = np.sqrt(np.mean(np.linalg.norm(p1c, axis=0) ** 2))

    s = rms_d0 / rms_d1
    P = s * np.eye(3).dot(R)

    t_final = P.dot(t1) - t0
    P = np.c_[P, t_final]

    return P


def face_nme(pre_pose_3d, gt_pose_3d, has_3d=None):
    assert pre_pose_3d.shape == gt_pose_3d.shape
    bs = pre_pose_3d.shape[0]

    bs_pre_points = pre_pose_3d.detach().cpu().numpy()
    bs_gt_points = gt_pose_3d.detach().cpu().numpy()

    pose_align_idx = [52, 55, 58, 61, 46, 84, 90]
    full_errors = []
    inner_face_errors = []
    for i in range(bs):
        if has_3d is not None:
            if has_3d[i] == 0.0:
                full_errors.append(0.0)
                inner_face_errors.append(0.0)
                continue

        pre_points = bs_pre_points[i]
        gt_points = bs_gt_points[i]

        Proj = estimate_pose(
            gt_points[pose_align_idx, :], pre_points[pose_align_idx, :]
        )
        pre_points = np.hstack((pre_points, np.ones((pre_points.shape[0], 1))))
        aligned_pre_points = Proj.dot(pre_points.T)
        aligned_pre_points = aligned_pre_points.T
        ## error
        # pdb.set_trace()
        errors = np.linalg.norm(aligned_pre_points - gt_points, axis=1)
        interocular_distance = np.linalg.norm(gt_points[74] - gt_points[77])
        full_errors.append(np.sum(errors) / (106 * interocular_distance))
        inner_face_errors.append(np.sum(errors[33:]) / (73 * interocular_distance))

    full_errors = float(np.mean(np.array(full_errors)))
    inner_face_errors = float(np.mean(np.array(inner_face_errors)))

    return full_errors, inner_face_errors



import cv2
import numpy as np

__all__ = ["plot_kps_2d", "imshow", "plot_fps", "plot_box", "plot_center_scale"]


def plot_kps_2d(img, kps_2d, color=(255, 0, 0), r=3):
    """在图片上绘制关节点"""

    if isinstance(kps_2d, list):
        kps_2d = np.array(kps_2d, dtype=np.float32)

    if len(kps_2d.shape) == 2:
        kps_2d = kps_2d[None, :, :]

    n, kps_num = kps_2d.shape[:2]
    for i in range(n):
        for j in range(kps_num):
            u, v = kps_2d[i, j, :2]
            cv2.circle(img, (int(u), int(v)), r, color, -1)


def imshow(title, img, wait_ms=10):
    """显示图像 Esc键退出"""

    if wait_ms <= 0:
        while True:
            cv2.imshow(title, img)
            if cv2.waitKey(20) == 27:  # Esc
                break
    else:
        cv2.imshow(title, img)
        cv2.waitKey(wait_ms)


def plot_fps(img, fps):
    """在图像上显示帧率"""

    H, W = img.shape[:2]
    cv2.putText(
        img,
        f"FPS:{fps:0.2f}",
        (W - 180, 50),
        cv2.FONT_HERSHEY_COMPLEX,
        1,
        (0, 0, 255),
        1,
    )


def plot_box(img, box, box_type="xywh", color=(255, 0, 0), thickness=4):
    """在图片上绘制矩形框"""

    if box_type == "xywh":
        x1, y1, w, h = box
        x2 = x1 + w
        y2 = y1 + h
    elif box_type == "xyxy":
        x1, y1, x2, y2 = box
    cv2.rectangle(
        img, (int(x1), int(y1)), (int(x2), int(y2)), color=color, thickness=thickness
    )


def plot_center_scale(img, center, scale, color=(0, 255, 0)):
    """在图片上绘制方框"""

    x = center[0] - scale / 2.0
    y = center[1] - scale / 2.0
    plot_box(img, [x, y, scale, scale], box_type="xywh", color=color)



import torch
import torch.nn as nn
# from timm.models.layers import DropPath # 版本报错


class Mlp(nn.Module):
    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        drop=0.0,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(
        self,
        dim,
        num_heads=8,
        qkv_bias=False,
        qk_scale=None,
        attn_drop=0.0,
        proj_drop=0.0,
    ):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads

        self.scale = qk_scale or head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x):
        B, N, C = x.shape
        qkv = (
            self.qkv(x)
            .reshape(B, N, 3, self.num_heads, C // self.num_heads)
            .permute(2, 0, 3, 1, 4)
        )
        q, k, v = qkv[0], qkv[1], qkv[2]

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class TransformerEncoderLayer(nn.Module):
    def __init__(
        self,
        dim,
        num_heads,
        mlp_ratio=4.0,
        qkv_bias=False,
        qk_scale=None,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            attn_drop=attn_drop,
            proj_drop=drop,
        )
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
        )

    def forward(self, x):
        x = x + self.drop_path(self.attn(self.norm1(x)))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x



import os
import os.path as osp
import pickle
import time
from typing import Union

import numpy as np

from ..tools.logger import Logger
from .track_base import BaseTrackObj
from .track_image import ImageTrackObj
from .track_number import NumberTrackObj
from .track_numpy import NumpyTrackObj


class Tracker(object):

    __filename = ""
    __start_time = None
    __is_open = False

    @staticmethod
    def trackfile(
        filename: str,
        open=True,
        clear=True,
    ) -> None:
        """设置track文件 *.pkl"""

        Tracker.__filename = filename
        Tracker.__start_time = time.time()

        path = osp.dirname(osp.abspath(filename))
        if not osp.exists(path):
            os.makedirs(path, exist_ok=True)

        Logger.warn(f"trackfile: {filename}")

        if clear and osp.exists(filename):
            os.remove(filename)

        if open:
            Tracker.open()

    @staticmethod
    def open() -> None:
        """开启记录功能"""
        Tracker.__is_open = True
        Logger.info("Tracker is opened!")

    @staticmethod
    def close() -> None:
        """关闭记录功能"""
        Tracker.__is_open = False
        Logger.info("Tracker is closed!")

    @staticmethod
    def is_open() -> bool:
        return Tracker.__is_open

    @staticmethod
    def track_obj(obj: BaseTrackObj, step=None) -> None:
        if not Tracker.is_open():
            return
        if not Tracker._check_file():
            return

        # 自动记录时间信息
        t = time.time()
        obj._time = t - Tracker.__start_time
        obj._abs_time = t
        obj._step = step
        with open(Tracker.__filename, "ab") as fw:
            pickle.dump(obj, fw)

    def track_img(img: np.ndarray, name: str = None, step: int = None) -> None:
        if Tracker.is_open():
            Tracker.track_obj(ImageTrackObj(img, name=name), step)

    def track_number(
        data: Union[float, int], name: str = None, step: int = None
    ) -> None:
        if Tracker.is_open():
            Tracker.track_obj(NumberTrackObj(data, name=name), step)

    def track_numpy(data: np.ndarray, name: str = None, step: int = None) -> None:
        if Tracker.is_open():
            Tracker.track_obj(NumpyTrackObj(data, name=name), step)

    @staticmethod
    def _check_file():
        if not Tracker.__filename:
            Logger.error("trackfile first!")
            return False
        return True



from .track_base import BaseTrackObj


class NumpyTrackObj(BaseTrackObj):
    def __init__(self, number, name=None) -> None:
        super().__init__()

        self.number = number
        self.set_name(name)

    def data(self):
        return self.number

    def show(self):
        print(self.data())



import logging
import os
import os.path as osp

import cv2
import logzero
from logzero import logger
import platform

# python3.8版本之后才支持 stacklevel
version = platform.python_version()
version = [int(a) for a in version.split(".")]
has_stacklevel = not (version[0] <= 3 and version[1] < 8)


class Logger(object):

    FOLDER = ""

    @staticmethod
    def debug(msg):
        if has_stacklevel:
            logger.debug(msg, stacklevel=2)
        else:
            logger.debug(msg)

    @staticmethod
    def info(msg, data=""):
        if has_stacklevel:
            logger.info(msg + str(data), stacklevel=2)
        else:
            logger.info(msg + str(data))

    @staticmethod
    def warn(msg, data=""):
        if has_stacklevel:
            logger.warning(msg + str(data), stacklevel=2)
        else:
            logger.warning(msg + str(data))

    @staticmethod
    def error(msg):
        if has_stacklevel:
            logger.error(msg, stacklevel=2)
        else:
            logger.error(msg)

    @staticmethod
    def logfile(filename, clear=False):
        """设置日志文件"""
        path = osp.dirname(filename)
        Logger.FOLDER = path
        if not osp.exists(path):
            os.makedirs(path, exist_ok=True)

        if has_stacklevel:
            logger.warning(f"logfile -> {filename}", stacklevel=2)
        else:
            logger.warning(f"logfile -> {filename}")
        if clear and osp.exists(filename):
            os.remove(filename)
        logzero.logfile(filename)

    @staticmethod
    def loglevel(level):
        """设置日志等级: debug, info, warn, error"""
        level = level.upper()
        logzero.loglevel(eval(f"logging.{level}"))

        if has_stacklevel:
            logger.warning(f"loglevel -> {level}", stacklevel=2)
        else:
            logger.warning(f"loglevel -> {level}")

    @staticmethod
    def logimg(filename, img):
        filename = osp.join(Logger.FOLDER, filename)
        cv2.imwrite(filename, img)
        Logger.warn(f"logimg: {filename}")


Logger.loglevel("debug")



import pickle
from typing import List
import os.path as osp
from .track_base import BaseTrackObj
from ..tools.logger import Logger


class TrackfileReader(object):
    def __init__(self, trackfile: str) -> None:
        assert osp.exists(trackfile)

        self.info_dict = dict()

        self.obj_list = []
        with open(trackfile, "rb") as fr:
            while True:
                try:
                    obj = pickle.load(fr)

                    name = obj._name
                    if name is not None:
                        if name not in self.info_dict:
                            self.info_dict[name] = []

                        self.info_dict[name].append(len(self.obj_list))
                        self.obj_list.append(obj)
                except EOFError:
                    break

    def get_obj(self, name: str, step: int) -> BaseTrackObj:
        if name in self.info_dict:
            k = self.info_dict[name][step]
            return self.obj_list[k]
        else:
            Logger.error(f"name={name} is bad")

    def get_length(self, name):
        return len(self.info_dict[name])

    def keys(self):
        return self.info_dict.keys()

def load_trackfile(trackfile) -> List[BaseTrackObj]:
    obj_list = []
    with open(trackfile, "rb") as fr:
        while True:
            try:
                obj = pickle.load(fr)
                obj_list.append(obj)
            except EOFError:
                break
    return obj_list


def analyzer(trackfile):
    obj_list = load_trackfile(trackfile)
    for obj in obj_list:
        obj.show()



def save_obj(filename, vertices, faces=None, colors=None):
    """保存obj文件
    vertices: shape Nx3
    faces: shape Mx3
    """

    if colors is not None:
        assert colors.shape == vertices.shape

    with open(filename, "w") as fw:
        if colors is None:
            for i in range(vertices.shape[0]):
                x, y, z = vertices[i]
                r, g, b = 0.5, 0.5, 0.5
                fw.write(f"v {x} {y} {z} {r} {g} {b}\n")
        else:
            for i in range(vertices.shape[0]):
                x, y, z = vertices[i]
                r, g, b = colors[i]
                fw.write(f"v {x} {y} {z} {r} {g} {b}\n")

        if faces is not None:
            for i in range(faces.shape[0]):
                a = faces[i][0]
                b = faces[i][1]
                c = faces[i][2]
                fw.write(f"f {a+1} {b+1} {c+1}\n")



smpl_path = ""
JOINT_REGRESSOR_TRAIN_EXTRA = smpl_path + 'J_regressor_extra.npy'
JOINT_REGRESSOR_H36M_correct = smpl_path + 'J_regressor_h36m_correct.npy'
SMPL_FILE = smpl_path + 'basicModel_neutral_lbs_10_207_0_v1.0.0.pkl'
SMPL_Male = smpl_path + 'basicModel_m_lbs_10_207_0_v1.0.0.pkl'
SMPL_Female = smpl_path + 'basicModel_f_lbs_10_207_0_v1.0.0.pkl'
SMPL_sampling_matrix = smpl_path + 'mesh_downsampling.npz'
MANO_FILE = smpl_path + 'MANO_RIGHT.pkl'
MANO_sampling_matrix = smpl_path + 'mano_downsampling.npz'

JOINTS_IDX = [
    8, 5, 29, 30, 4, 7, 21, 19, 17, 16, 18, 20, 31, 32, 33, 34, 35, 36, 37, 24,
    26, 25, 28, 27
]
"""
We follow the body joint definition, loss functions, and evaluation metrics from 
open source project GraphCMR (https://github.com/nkolot/GraphCMR/)

Each dataset uses different sets of joints.
We use a superset of 24 joints such that we include all joints from every dataset.
If a dataset doesn't provide annotations for a specific joint, we simply ignore it.
The joints used here are:
"""

SMPL_J24_IDX = {
    "right_ankle": 0,
    "right_knee": 1,
    "right_hip": 2,
    "left_hip": 3,
    "left_knee": 4,
    "left_ankle": 5,
    "right_wrist": 6,
    "right_elbow": 7,
    "right_shoulder": 8,
    "left_shoulder": 9,
    "left_elbow": 10,
    "left_wrist": 11,
    "neck": 12,
    "top_of_head": 13,
    "pelvis": 14,
    "thorax": 15,
    "spine": 16,
    "jaw": 17,
    "head": 18,
    "nose": 19,
    "left_eye": 20,
    "right_eye": 21,
    "left_ear": 22,
    "right_ear": 23,
}
#  [14, 3, 4, 5, 2, 1, 0, 16, 12, 17, 18, 9, 10, 11, 8, 7, 6]
COMMON_J12_IDX = {
    "right_ankle": 0,
    "right_knee": 1,
    "right_hip": 2,
    "left_hip": 3,
    "left_knee": 4,
    "left_ankle": 5,
    "right_wrist": 6,
    "right_elbow": 7,
    "right_shoulder": 8,
    "left_shoulder": 9,
    "left_elbow": 10,
    "left_wrist": 11,
}

BODY_EDGE_LIST = [
    ["right_ankle", "right_knee"],
    ["right_knee", "right_hip"],
    ["right_hip", "right_shoulder"],
    ["right_shoulder", "right_elbow"],
    ["right_elbow", "right_wrist"],

    ["left_ankle", "left_knee"],
    ["left_knee", "left_hip"],
    ["left_hip", "left_shoulder"],
    ["left_shoulder", "left_elbow"],
    ["left_elbow", "left_wrist"],
]

J14_IDX = {
    "right_ankle": 0,
    "right_knee": 1,
    "right_hip": 2,
    "left_hip": 3,
    "left_knee": 4,
    "left_ankle": 5,
    "right_wrist": 6,
    "right_elbow": 7,
    "right_shoulder": 8,
    "left_shoulder": 9,
    "left_elbow": 10,
    "left_wrist": 11,
    "neck": 12,
    "head": 13,
}
COCO_J17_IDX = {
    "nose": 0,
    "left_eye": 1,
    "right_eye": 2,
    "left_ear": 3,
    "right_ear": 4,
    "left_shoulder": 5,
    "right_shoulder": 6,
    "left_elbow": 7,
    "right_elbow": 8,
    "left_wrist": 9,
    "right_wrist": 10,
    "left_hip": 11,
    "right_hip": 12,
    "left_knee": 13,
    "right_knee": 14,
    "left_ankle": 15,
    "right_ankle": 16,
}

COCO_Skeleton = [[16, 14], [14, 12], [17, 15], [15, 13], [12, 13], [6, 12], [7, 13], [6, 7], [6, 8], [7, 9], [8, 10], [9, 11], [2, 3], [1, 2], [1, 3], [2, 4], [3, 5], [4, 6], [5, 7]]


MPII_J16_IDX = {
    "right_ankle": 0,
    "right_knee": 1,
    "right_hip": 2,
    "left_hip": 3,
    "left_knee": 4,
    "left_ankle": 5,
    "pelvis": 6,
    "thorax": 7,
    "neck": 8,
    "top_of_head": 9,
    "right_wrist": 10,
    "right_elbow": 11,
    "right_shoulder": 12,
    "left_shoulder": 13,
    "left_elbow": 14,
    "left_wrist": 15,
}
H36M_J17_IDX = {
    "pelvis": 0,
    "right_hip": 1,
    "right_knee": 2,
    "right_ankle": 3,
    "left_hip": 4,
    "left_knee": 5,
    "left_ankle": 6,
    "torso": 7,
    "neck": 8,
    "nose": 9,
    "head": 10,
    "left_shoulder": 11,
    "left_elbow": 12,
    "left_wrist": 13,
    "right_shoulder": 14,
    "right_elbow": 15,
    "right_wrist": 16,
}

def convert_idx(from_idx, to_idx):
    index_list = []
    for i, name in enumerate(to_idx):
        assert to_idx[name] == i
        index_list.append(from_idx[name])
    return index_list

SMPL_TO_COCO = convert_idx(SMPL_J24_IDX, COCO_J17_IDX)
SMPL_TO_MPII = convert_idx(SMPL_J24_IDX, MPII_J16_IDX)
SMPL_TO_J14 = convert_idx(SMPL_J24_IDX, J14_IDX)
SMPL_TO_J12 = convert_idx(SMPL_J24_IDX, COMMON_J12_IDX)
H36M_TO_J14 = convert_idx(H36M_J17_IDX, J14_IDX)
H36M_TO_J12 = convert_idx(H36M_J17_IDX, COMMON_J12_IDX)
COCO_TO_J12 = convert_idx(COCO_J17_IDX, COMMON_J12_IDX)

# print("SMPL_TO_J12", SMPL_TO_J12)
# print("COCO_TO_J12", COCO_TO_J12)
# COCO_TO_J14 = convert_idx(COCO_J17_IDX, J14_IDX)
# MPII_J16_NAME = ("R_Ankle", "R_Knee", "R_Hip", "L_Hip", "L_Knee", "L_Ankle",
#                  "Pelvis", "Thorax", "Neck", "Top_of_Head", "R_Wrist",
#                  "R_Elbow", "R_Shoulder", "L_Shoulder", "L_Elbow", "L_Wrist")
# J24_TO_MPII = [J24_NAME.index(name) for name in MPII_J16_NAME]

H36M_J17_NAME = ('Pelvis', 'R_Hip', 'R_Knee', 'R_Ankle', 'L_Hip', 'L_Knee',
                 'L_Ankle', 'Torso', 'Neck', 'Nose', 'Head', 'L_Shoulder',
                 'L_Elbow', 'L_Wrist', 'R_Shoulder', 'R_Elbow', 'R_Wrist')
J24_TO_J14 = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 18]
H36M_J17_TO_J14 = [3, 2, 1, 4, 5, 6, 16, 15, 14, 11, 12, 13, 8, 10]

J14_NAME = [H36M_J17_NAME[idx] for idx in H36M_J17_TO_J14]
# print("J14_NAME", J14_NAME)
"""
We follow the hand joint definition and mesh topology from 
open source project Manopth (https://github.com/hassony2/manopth)

The hand joints used here are:
"""
J_NAME = ('Wrist', 'Thumb_1', 'Thumb_2', 'Thumb_3', 'Thumb_4', 'Index_1',
          'Index_2', 'Index_3', 'Index_4', 'Middle_1', 'Middle_2', 'Middle_3',
          'Middle_4', 'Ring_1', 'Ring_2', 'Ring_3', 'Ring_4', 'Pinky_1',
          'Pinky_2', 'Pinky_3', 'Pinky_4')
ROOT_INDEX = 0



import torch.nn as nn


class KeypointLoss(object):
    def __init__(self, use_mse=False) -> None:
        super().__init__()
        if use_mse:
            self.criterion_2d = nn.MSELoss(reduction='none')
            self.criterion_3d = nn.MSELoss(reduction='none')
        else:
            self.criterion_2d = nn.L1Loss()
            self.criterion_3d = nn.L1Loss()
        self.criterion_vertices = nn.L1Loss()

    def cal_pose_2d_loss(self, pred_pose_2d, gt_pose_2d):
        """
        Compute 2D reprojection loss if 2D keypoint annotations are available.
        The confidence (conf) is binary and indicates whether the keypoints exist or not.
        """

        conf = gt_pose_2d[:, :, -1:]
        return (conf * self.criterion_2d(pred_pose_2d, gt_pose_2d[:, :, :-1])).mean()

    def cal_pose_3d_loss(self, pred_pose_3d, gt_pose_3d, has_pose_3d):
        """
        Compute 3D keypoint loss if 3D keypoint annotations are available.
        """
        gt_pose_3d = gt_pose_3d[has_pose_3d == 1]
        conf = gt_pose_3d[:, :, -1:]
        return (
            conf
            * self.criterion_3d(pred_pose_3d[has_pose_3d == 1], gt_pose_3d[:, :, :-1])
        ).mean()

    def cal_vertices_loss(self, pred_vertices, gt_vertices, has_smpl):
        """
        Compute per-vertex loss if vertex annotations are available.
        """
        return self.criterion_vertices(
            pred_vertices[has_smpl == 1], gt_vertices[has_smpl == 1]
        )
